* form_run, a form_do replacement - version 3.0 of 2 September 1991
* Copyright 1991 by Bill Aycock 76703,4061
* If you use this routine, please mention it in your documentation!
*
*NOTE: In this version, if the procedure is running as a desk accessory,
*  then the extended-object-type functions (exiting or highlighting an
*  object when the mouse pointer passes over it) and the dialog exit when
*  the user clicks off the form do not function correctly... if the user
*  _does_ click off the form, these functions won't work until the user
*  clicks on one of the dialog's buttons. All other functions like the
*  command keys, key equivalents, keypress substitutions, and extended
*  text editing features will still function as documented.
*
*-----------------------------------------------------------------------*
*in:	int_in(0) = first editable object number, or 0 if none
*		(identical to form_do parameter)
*	int_in(1) = extras flag - bits mean:
*		bit 0 set = exit if user clicks outside form
*		bit 1 set = wrap editable fields first->last, last->first
*		bit 2 set = shift-tab acts like up-arrow (prev edit obj)
*		bit 3 set = enable Extended Character Entry (see below)
*	addr_in(0) = address of tree
*		(identical to form_do parameter)
*	addr_in(1) = address of command-key list, or 0 if no list...
*		List is null-terminated list of words, each entry of
*		which is a key code (such as returned by evnt_keybd)
*		indicating a keypress to be treated as an exit object.
*	    ALTERNATE FORM:
*		An entry in the list may consist of TWO WORDS.
*		The FIRST word must be a MASK of shift-key bits such as
*		would be returned from a kb_shift call. The bits mean:
*			bit 0 set = right shift key pressed
*			bit 1 set = left shift key pressed
*			bit 2 set = control key pressed
*			bit 3 set = alternate key pressed
*			bits 4-15 = reserved, MUST BE ZERO
*		The SECOND word of the entry must contain the SCAN CODE
*		for the keypress in the high byte, and the low byte of
*		the word, normally the ASCII value, MUST BE ZERO.
*	addr_in(2) = address of key-equivalent list, or 0 if no list...
*		List is null-terminated list of word pairs. In each pair,
*		the first word is a key code (as above) and the second
*		is an object number. When the keycode is pressed, this
*		routine will translate it into a click on the specified
*		object. If bit 15 of the object number is set, form_run
*		will signal a DOUBLE-click on the object.
*	    ALTERNATE FORM:
*		An entry in the list may consist of THREE WORDS.
*		The FIRST word must be a MASK of shift-key bits such as
*		would be returned from a kb_shift call. The bits mean:
*			bit 0 set = right shift key pressed
*			bit 1 set = left shift key pressed
*			bit 2 set = control key pressed
*			bit 3 set = alternate key pressed
*			bits 4-15 = reserved, MUST BE ZERO
*		The SECOND word of the entry must contain the SCAN CODE
*		for the keypress in the high byte, and the low byte of
*		the word, normally the ASCII value, MUST BE ZERO.
*		The THIRD word of the entry should contain the desired
*		object number as above.
*	addr_in(3) = address of key-substitute list, or 0 if no list...
*		List is null-terminated list of word pairs. In each pair,
*		the first word is a key code (as above) for a user
*		keypress, and the second is a key code that will be
*		substituted for the actual keypress.
*	    ALTERNATE FORM:
*		An entry in the list may consist of THREE WORDS.
*		The FIRST word must be a MASK of shift-key bits such as
*		would be returned from a kb_shift call. The bits mean:
*			bit 0 set = right shift key pressed
*			bit 1 set = left shift key pressed
*			bit 2 set = control key pressed
*			bit 3 set = alternate key pressed
*			bits 4-15 = reserved, MUST BE ZERO
*		The SECOND word of the entry must contain the SCAN CODE
*		for the keypress in the high byte, and the low byte of
*		the word, normally the ASCII value, MUST BE ZERO.
*		The THIRD word of the entry should contain the desired
*		substitute keypress as above.
* NOTE: The alternate forms of the three keypress lists give you even
*	more control over your program's response to keyboard events.
*	Use the alternate forms to respond to unusual key combinations
*	such as shift-alternate keys, shift-control keys, etc.
*-----------------------------------------------------------------------*
*out:	int_out(0) = exit object (also returned in d0)...
*	returned word is the object number of selected exit object
*	if result is zero:
*		user clicked outside form, and bit 0 of Extras flag
*		from int_in(1) above was set
*	if bit 15 of result is set:
*		exit object was touch-exit, and was double-clicked
*		(just like normal touchexit object handling)
*	if bit 14 of result is set:
*		exit object was a command key, and low byte of result
*		is the command key's position in list (starting at zero)
*	if bit 13 of result is set:
*		exit object was _not_ clicked; bit 1 of the object's
*		extended-type byte was set, and the mouse pointer
*		passed over the object (see Extended Object Types below)
*-----------------------------------------------------------------------*
*Extended Text Editing:
*	Several convenient extensions for editable text fields are always
*	available when using form_run instead of form_do...
*
*	* shift-left-arrow moves the text cursor to the first character
*		of the current editable text field
*	* shift-right-arrow moves the cursor to the field's last character
*	* clr/home moves the text cursor to the first editable field
*		on the form (excluding hidden objects)
*	* shift-clr/home moves the cursor to the last editable field
*		on the form (excluding hidden objects)
*
*Keypresses are processed in this order:
*	(1) command-key exit keys	- see addr_in(1)
*	(2) keypress/object equivalents	- see addr_in(2)
*	(3) keypress substitution	- see addr_in(3)
*	(4) Extended Character Entry	- see int_in(1)
*	(5) all other ("normal") keypresses
*-----------------------------------------------------------------------*
*Extended Character Entry:
*	If bit 3 of the Extras flag int_in(1) is set, the user can enter
*	any character into an editable string by _holding down_ the Alt
*	key, pressing up to three _decimal_ digits ON THE NUMERIC KEYPAD,
*	then _releasing_ the Alt key.  A character entered this way is
*	checked for validity just like any normal keypress.
*
*	While waiting for keypad digit entry:
*		* mouse cursor turns into 'busy bee' shape
*		* all keys other than keypad digits 0-9 are ignored
*		* numbers greater than 255 are ignored
*-----------------------------------------------------------------------*
*Extended Object Types:
*	The ob_type word for each object shows what type of object it
*	is.  The high byte is reserved for extended types.  For this
*	routine, two bits in the extended-type byte are used:
*
*	if bit 0 is set:
*		object will highlight when the mouse passes over it, and
*		unhighlight when mouse moves away
*		(only affects non-selected objects)
*	if bit 1 is set:
*		when mouse passes over the object, form_run exits and
*		returns the object number with bit 13 set
*		(similar to handling of double-clicked touchexit objects)
*-----------------------------------------------------------------------*

	section text

PASSEXIT	equ	13	;result bit showing passover exit
CMDKEYEXIT	equ	14	;result bit showing cmd-key exit
DCLICKEXIT	equ	15	;result bit showing double-click exit

form_run: movem.l	d1-d7/a0-a6,-(sp)	;save regs
	bsr		fr_init			;save/init stuff
	tst.w		fr_editobj		;init edit if needed
	beq.s		.multi
	objc_edit	fr_treeaddr,fr_editobj,#0,fr_editidx,#ED_INIT
	move.w		int_out+2,fr_editidx	;save new index

*multi parameters:
*	events
*	clicks=2, but.mask=Left, but.state=LdownRup
*	m1.flag-x-y-w-h
*	m2.f-x-y-w-h (no args, no MOVE.Ws)
*	buf,timer (no timer arg, no MOVEs)
.multi:	clr.b		fr_exitnow		;clear exit flag
	evnt_multi	#MU_KEYBD+MU_BUTTON+MU_M1,
&			#2,#1,#1,
&			#1,fr_mousex,fr_mousey,#1,#1,
&			,,,,,
&			#fr_msgbuf
	move.w	int_out,fr_happened		;save results now...
	move.l	int_out+2,fr_mousex		;mouse x, mouse y
	move.l	int_out+6,fr_button		;button, shift
	move.l	int_out+10,fr_keycode		;keycode, clicked

.notkey: btst.b	#1,fr_happened+1	;button event?
	beq.s	.notbut
	bsr	fr_butn
.notbut: btst.b	#2,fr_happened+1	;mouse movement?
	beq.s	.notm1
	bsr	fr_m1
.notm1:	btst.b	#0,fr_happened+1	;was this a keyboard event?
	beq.s	.stay
	bsr	fr_keybd

.stay:	tst.b	fr_exitnow		;exit flag say time to leave?
	beq	.multi			;nope, check next event
.exnow:	tst.w	fr_editobj		;end edit if needed
	beq.s	.noed
	objc_edit	fr_treeaddr,fr_editobj,#0,fr_editidx,#ED_END
.noed:	tst.w	fr_hilited		;have leftover hilited object?
	beq.s	.exit
	bsr	fr_unhilite
.exit:	move.w	fr_result,d0		;grab result now
	move.w	d0,int_out
	movem.l	(sp)+,d1-d7/a0-a6	;restore regs
	rts
*----------


*save parms, init any needed variables
*----------
fr_init: move.l	int_in,fr_editobj	;save parms
	move.l	addr_in,fr_treeaddr	;tree address
	move.l	addr_in+4,fr_cmdkeys	;command keys
	move.l	addr_in+8,fr_equkeys	;key equivalents
	move.l	addr_in+12,fr_subkeys	;key substitutions
	clr.w	fr_oldobj
	clr.w	fr_hilited
	clr.l	fr_mousex		;(clears 2 words, x & y!)
	rts
*----------


*handle keyboard events
*----------
fr_keybd: movea.l	fr_cmdkeys,a0		;get ptr to cmd keys
	cmpa.l		#0,a0			;any list?
	beq.s		.ckequ			;nope!
	move.w		fr_keycode,d1		;grab the keypress
	moveq		#0,d0			;init our counter
.kloop:	move.w		(a0)+,d2		;get cmdkey from list
	beq.s		.ckequ			;list ends, chk for others!
	cmpi.w		#$ff,d2			;no scancode in entry?
	bgt.s		.kloo2			;is one, normal key
	cmp.w		fr_shift,d2		;shift bits match?
	beq.s		.kchk			;yup, now check scan code!
	addq.l		#2,a0			;nope, skip scancode word
	bra.s		.kloox			;go try next keycode
.kchk:	move.w		(a0)+,d2		;get scancode word now
	move.w		d1,d3			;make copy of keycode
	clr.b		d3			;clear low BYTE
	cmp.w		d3,d2			;matches our scancode?
	beq.s		.match			;yes, show result!
	bra.s		.kloox			;no, go try next keycode
.kloo2:	cmp.w		d1,d2			;matches our keypress?
	beq.s		.match			;yes, show result!
.kloox:	addq.w		#1,d0			;increment counter
	bra.s		.kloop			;go try next keycode
.match:	move.b		#1,fr_exitnow		;set our exit flag
	bset.l		#CMDKEYEXIT,d0		;set result's cmd key bit
	move.w		d0,fr_result		;show the result
	rts					;and quit!
.ckequ:	movea.l		fr_equkeys,a0		;get ptr to equivalents
	cmpa.l		#0,a0			;any list?
	beq.s		.cksub			;nope!
	move.w		fr_keycode,d1		;grab the keypress
.eloop:	move.w		(a0)+,d2		;get equkey from list
	beq.s		.cksub			;list ends, chk for others!
	cmpi.w		#$ff,d2			;no scancode in entry?
	bgt.s		.eloo2			;is one, normal key
	cmp.w		fr_shift,d2		;shift bits match?
	beq.s		.echk			;yup, now check scan code!
	addq.l		#2,a0			;nope, skip scancode word
	bra.s		.eloox			;go try next keycode
.echk:	move.w		(a0)+,d2		;get scancode word now
	move.w		d1,d3			;make copy of keycode
	clr.b		d3			;clear low BYTE
	cmp.w		d3,d2			;matches our scancode?
	beq.s		.equiv			;yes, show result!
	bra.s		.eloox			;no, go try next keycode
.eloo2:	cmp.w		d1,d2			;matches our keypress?
	beq.s		.equiv			;yes, make it a click!
.eloox:	addq.l		#2,a0			;skip object number
	bra.s		.eloop			;go try next keycode
.equiv:	move.w		(a0),d0			;grab the object #
	move.w		#1,fr_clicked		;show single click
	btst.l		#15,d0			;obj's high bit set?
	beq.s		.clik			;nope, use 1 click
	bclr.l		#15,d0			;clear hi bit now
	move.w		#2,fr_clicked		;show a DOUBLEclick!
.clik:	bra		fr_obj			;jump (not jsr!) to handler
.cksub:	movea.l		fr_subkeys,a0		;get ptr to key substitutes
	cmpa.l		#0,a0			;any list?
	beq.s		.ckret			;nope!
	move.w		fr_keycode,d1		;grab the keypress
.sloop:	move.w		(a0)+,d2		;get subkey from list
	beq.s		.ckret			;list ends, chk for others!
	cmpi.w		#$ff,d2			;no scancode in entry?
	bgt.s		.sloo2			;is one, normal key
	cmp.w		fr_shift,d2		;shift bits match?
	beq.s		.schk			;yup, now check scan code!
	addq.l		#2,a0			;nope, skip scancode word
	bra.s		.sloox			;go try next keycode
.schk:	move.w		(a0)+,d2		;get scancode word now
	move.w		d1,d3			;make copy of keycode
	clr.b		d3			;clear low BYTE
	cmp.w		d3,d2			;matches our scancode?
	beq.s		.subs			;yes, show result!
	bra.s		.sloox			;no, go try next keycode
.sloo2:	cmp.w		d1,d2			;matches our keypress?
	beq.s		.subs			;yes, substitute new keypress!
.sloox:	addq.l		#2,a0			;skip replacement keypress
	bra.s		.sloop			;go try next keycode
.subs:	move.w		(a0),fr_keycode		;replace keypress & continue!
.ckret:	cmpi.b		#$0d,fr_keycode+1	;chk for return/enter key
	bne.s		.cked			;not one, look for ed chars
	bsr		fr_treewalk		;go find 1st default
	move.w		d3,fr_result		;get ready to return it
	beq		.exit			;if =0 then no default!
	move.w		fr_result,d1		;show which object
	moveq		#1,d2			;flag to select it
	bsr		fr_redraw		;redraw it
	move.b		#1,fr_exitnow		;signal exit-now
	bra		.exit			;quit but don't edit char
.cked:	tst.w		fr_editobj		;any editable object?
	beq		.exit			;nope, all done!
	cmpi.w		#$4700,fr_keycode	;home (1st edobj)?
	bne.s		.shome
	bsr		fr_treewalk		;get object stats
	move.w		d4,fr_work		;show 1st edobj
	bsr		fr_chged		;and switch to it!
	bra		.exit
.shome:	cmpi.w		#$4737,fr_keycode	;shift-home (last edobj)?
	bne.s		.sleft
	bsr		fr_treewalk		;get object stats
	move.w		d7,fr_work		;show last edobj
	bsr		fr_chged		;and switch to it!
	bra		.exit
.sleft:	cmpi.w		#$4b34,fr_keycode	;shift-left (1st char)?
	bne		.srite
	objc_edit	fr_treeaddr,fr_editobj,#0,fr_editidx,#ED_END
	movea.l		fr_treeaddr,a3		;get obj's address
	move.w		fr_editobj,d3
	mulu		#24,d3
	lea		(a3,d3.L),a3
	movea.l		ob_spec(a3),a3		;point to tedinfo
	movea.l		(a3),a3			;point to actual text
	move.b		(a3),d3			;save 1st character
	clr.b		(a3)			;then zero it out
	objc_edit	fr_treeaddr,fr_editobj,#0,#0,#ED_INIT
	move.b		d3,(a3)			;restore 1st character
	clr.w		fr_editidx		;point to pos 0 and
	move.w		#$4b00,fr_keycode	;use an L-arrow to
	bra.s		.obed			;redraw text <phew!>
.srite:	cmpi.w		#$4d36,fr_keycode	;shift-right (last char)?
	bne.s		.ckalt
	move.w		fr_editobj,fr_work	;set up for edit call...
	bsr		fr_yes			;move crsr to last char!
	bra.s		.exit			;all done!
.ckalt:	btst.b		#3,fr_shift+1		;check alt key state
	beq.s		.noalt			;skip if not pressed
	bsr		fr_altkey		;check for alt-numbers...
.noalt:	bsr		fr_objkey		;check for new-ed-obj keys
	tst.w		d0			;chg object?
	bne.s		.exit			;yup, skip char edit!
.obed:	objc_edit	fr_treeaddr,fr_editobj,
&			fr_keycode,fr_editidx,#ED_CHAR
	move.w		int_out+2,fr_editidx	;save new index
.exit:	rts
*----------


*handle arrows/tabs/return to select new edit object
*----------
fr_objkey: cmpi.w	#$4800,fr_keycode	;up arrow?
	beq.s		.prev			;go find prev ed obj
.nup:	cmpi.w		#$5000,fr_keycode	;down arrow?
	beq.s		.next			;go find next ed obj
.ndown:	cmpi.w		#$0f09,fr_keycode	;real tab key?
	beq.s		.tgo			;yup, go check shift!
	moveq		#0,d0			;show no editobj change
	bra		.exit			;not special key, exit!
.tgo:	btst.b		#2,fr_extras+1		;shift-tab goes backward?
	beq.s		.next			;nope, handle normal
	move.w		fr_shift,d1		;tab shifted or not?
	andi.w		#%11,d1			;test only L & R shift keys
	beq.s		.next			;not shifted, do next obj
.prev:	bsr		fr_treewalk		;go find prev edit obj
	move.w		d5,fr_work		;use prev as candidate
	bne.s		.newed			;was a prev, use it!
	btst.b		#1,fr_extras+1		;wrap flag on?
	beq.s		.newed			;nope, leave as is!
	move.w		d7,fr_work		;use last edit object
	bra.s		.newed
*d4=1sted,d5=preved,d6=nexted,d7=lasted;fr_extras/bit1=wrap on
.next:	bsr		fr_treewalk		;go find next edit obj
	move.w		d6,fr_work		;use next as candidate
	bne.s		.newed			;was a next, use it!
	btst.b		#1,fr_extras+1		;wrap flag on?
	beq.s		.newed			;nope, leave as is!
	move.w		d4,fr_work		;use first edit object
.newed:	tst.w		fr_work			;any new-edit-obj candidate?
	beq.s		.done			;nope, quit w/o chg or edit
	bsr		fr_chged		;change edit obj if needed
.done:	move.w		#$ffff,d0		;don't pass chr to ed routine
.exit:	rts
*----------


*walk entire tree and pass back:
* 1st default	d3
* 1st editable	d4
* prev editable	d5
* next editable	d6
* last editable	d7
*----------
fr_treewalk:	movea.l	fr_treeaddr,a0	;a0=current obj address
	movea.l	a0,a1			;a1=tree address
*					;a2=work address
*				;fr_editobj=entry edit object#
*					;d1=work obj
	moveq	#0,d2			;d2=current object#
	moveq	#0,d3			;d3=1st default obj
	moveq	#0,d4			;d4=1st editable obj
	moveq	#0,d5			;d5=prev editable obj
	moveq	#0,d6			;d6=next editable obj
	moveq	#0,d7			;d7=last editable obj
	clr.b	fr_passed		;flag, 1=passed current ed obj
	bsr.s	fr_chk			;go check for new stats
*1. look at ob_head. if not -1, make it current obj & goto 1.
.s1:	move.w	ob_head(a0),d0		;grab ob_head
	cmpi.w	#-1,d0			;is it -1?
	beq.s	.s2			;yep, goto step 2.
*1a. see if obj is hidden! if so, skip the children (pretend ob_head=-1)
*	and go right to ob_next part (step2) instead.
	btst.b	#7,ob_flags+1(a0)	;hidden bit set?
	bne.s	.s2			;yep, skip this subtree!
.s1a:	move.w	d0,d2			;make it current object....
	mulu	#24,d0			;find its address...
	lea	(a1,d0.L),a0
	bsr.s	fr_chk			;go check for new stats
	bra.s	.s1			;go back to step1.
*2. look at ob_next. if =-1, back at root (visited all objs).
.s2:	move.w	ob_next(a0),d0		;grab ob_next
	cmpi.w	#-1,d0			;is it -1?
	beq.s	.exit			;yup, seen whole tree!
*3. (ob_next not= -1). see if object ob_next pts to contains # of current
*	obj in its ob_tail field. if so, we're back at the parent of
*	current obj, so make it the new current obj and goto step2.
.s3:	move.w	d0,d1			;get ob_next address into a2
	mulu	#24,d1
	lea	(a1,d1.L),a2
	cmp.w	ob_tail(a2),d2		;ob_next's ob_tail= current?
	bne.s	.s4			;nope, continue
	move.w	d0,d2			;yup, make it current obj...
	movea.l	a2,a0
	bsr.s	fr_chk			;go check for new stats
	bra.s	.s2			;and goto step2.
*4. if ob_next not -1, and doesn't fit #3, then it pts to a sibling
*	we haven't visited. make it current obj and goto #1.
.s4:	move.w	d0,d2			;new sibling, make it current obj...
	movea.l	a2,a0
	bsr.s	fr_chk			;go check for new stats
	bra.s	.s1			;and goto step1.
.exit:	rts
*----------


*routine to check for default/editable objects (skips hidden objs!)
*a0=current object's address, d2.w=this object #
*----------
fr_chk:	btst.b	#7,ob_flags+1(a0)		;hidden bit set?
	beq.s	.go				;nope, continue
	rts					;yep, skip it!
.go:	btst.b	#1,ob_flags+1(a0)		;default bit set?
	beq.s	.s2				;nope, chk next part
	tst.w	d3				;got default yet?
	bne.s	.s2				;yup, skip this.
	move.w	d2,d3				;save 1st default obj.
.s2:	tst.w	fr_editobj			;any editing?
	beq.s	.exit				;nope, skip rest!
.s3:	cmp.w	fr_editobj,d2			;this is current edobj?
	bne.s	.s4				;nope, continue
	move.b	#1,fr_passed			;yup, set flag
.s4:	btst.b	#3,ob_flags+1(a0)		;editable obj?
	beq.s	.exit				;nope, skip rest!
	move.w	d2,d7				;save for last-edobj
	tst.w	d4				;had 1st already?
	bne.s	.s5				;yup, skip this.
	move.w	d2,d4				;save 1st-edobj.
.s5:	cmp.w	fr_editobj,d2			;current edobj?
	beq.s	.exit				;yup, skip rest!
	tst.b	fr_passed			;passed current edobj?
	bne.s	.s6				;yup, skip this.
	move.w	d2,d5				;save for prev-edobj
	bra.s	.exit				;and skip next test.
.s6:	tst.w	d6				;already have next edobj?
	bne.s	.exit				;yup, all done!
	move.w	d2,d6				;save next-edobj
.exit:	rts
*----------


*handle alt-keypad character entry
*----------
fr_altkey:	btst.b	#3,fr_extras+1	;extended entry enabled?
	bne.s		.go		;yup, go for it!
.quit:	rts
.go:	move.b		fr_keycode+1,d0	;see if altpadnum
	cmpi.b		#$30,d0		;too small?
	blt.s		.quit		;yup, don't start
	cmpi.b		#$39,d0		;too big?
	bgt.s		.quit		;yup, don't start
	clr.l		fr_altstr	;clear result...
	move.b		d0,fr_altstr	;save partial result
	lea		fr_altstr+1,a3	;point to next char spot
	graf_mouse	#HOURGLASS	;turn off mouse
.loop:	move.w		#-1,-(sp)	;chk alt key...
	move.w		#Kbshift,-(sp)
	trap		#BIOS
	addq.l		#4,sp
	btst.l		#3,d0		;alt key pressed?
	beq.s		.done		;nope, send results!
	move.w		#CON,-(sp)	;chk if char waiting...
	move.w		#Bconstat,-(sp)
	trap		#BIOS
	addq.l		#4,sp
	tst.w		d0		;key waiting?
	beq.s		.loop		;nope
	move.w		#CON,-(sp)	;get character...
	move.w		#Bconin,-(sp)
	trap		#BIOS
	addq.l		#4,sp
	cmpi.b		#$30,d0		;too small?
	blt.s		.loop		;yup
	cmpi.b		#$39,d0		;too big?
	bgt.s		.loop		;yup
	move.b		d0,(a3)+	;save the digit
	cmpa.l		#fr_altstr+3,a3	;got 3 yet?
	bne.s		.loop		;nope, get another
.done:	lea		fr_altstr,a0	;point to string
	clr.l		d0		;clear accumulator
.cvt:	tst.b		(a0)		;finished?
	beq.s		.got		;yup, go on
	lsl.l		#1,d0		;mult by 2
	move.l		d0,d1		;save this
	lsl.l		#2,d0		;mult by 4 (by 8 now)
	add.l		d1,d0		;now it's by 10
	move.b		(a0)+,d1	;fetch digit
	andi.l		#$0f,d1		;save low byte (1 digit) only
	add.l		d1,d0		;add into accum
	bra.s		.cvt		;go do another digit
.got:	move.w		d0,fr_keycode	;put result in place
	cmpi.w		#$ff,d0		;was code too big?
	ble.s		.exit		;nope, send it on!
	clr.w		fr_keycode	;clear bad keypress
.exit:	graf_mouse	#ARROW		;turn on mouse again
	rts
*----------


*handle mouse button events
*----------
fr_butn: objc_find	fr_treeaddr,#0,#10,fr_mousex,fr_mousey
	tst.w		d0			;any object there?
	bpl.s		fr_obj			;yup, go check it!
	btst.b		#0,fr_extras+1		;exit if click off form?
	beq.s		.noobj			;nope, bit not set, ding!
	moveq		#0,d0			;yep, show the result
	move.b		#1,fr_exitnow		;show we wanna quit
	bra.s		.exit			;and quit!
.noobj:	move.w		#7,-(sp)		;ding!
	move.w		#CON,-(sp)
	move.w		#Bconout,-(sp)
	trap		#BIOS
	addq.l		#6,sp
.exit:	move.w		d0,fr_result		;show results
	rts
*----------


*handle mouse-click on an object (from fr_butn or fr_keybd!)
*object # passed in d0.W
*----------
fr_obj:	movea.l		fr_treeaddr,a0		;get pointer to obj
	move.w		d0,fr_newobj		;save obj#
	mulu		#24,d0			;get pointer to obj
	lea		(a0,d0.L),a0
	btst.b		#3,ob_flags+1(a0)	;was it editable?
	beq.s		.noed			;nope
	move.w		fr_newobj,fr_work	;show new candidate
	bsr		fr_chged		;and change if needed
.noed:	objc_offset	fr_treeaddr,fr_newobj	;must do an objc_find for
	move.l		int_out+2,fr_x		; form_butn to work!!!!!
	objc_find	fr_treeaddr,#0,#10,fr_x,fr_y
	form_butn	fr_treeaddr,fr_newobj,fr_clicked
	move.w		int_out+2,d0		;put result in place
	clr.b		fr_exitnow		;don't exit yet though
	tst.w		int_out			;exit object flagged?
	bne.s		.exit			;nope, continue
.quit:	move.b		#1,fr_exitnow		;show we wanna quit
.exit:	move.w		d0,fr_result		;show results
	rts
*----------


*fr_chged: change the edit object if different (new one passed in fr_work)
*fr_yes: move cursor to last character!
*----------
fr_chged:	move.w	fr_editobj,d1		;grab current
	cmp.w		fr_work,d1		;same as new?
	beq.s		fr_ex			;yup, skip this!
fr_yes:	objc_edit	fr_treeaddr,fr_editobj,#0,fr_editidx,#ED_END
	move.w		fr_work,fr_editobj
	objc_edit	fr_treeaddr,fr_editobj,#0,#0,#ED_INIT
	move.w		int_out+2,fr_editidx	;save new index
fr_ex:	rts
*----------


*handle mouse movement events
*----------
fr_m1:	objc_find	fr_treeaddr,#0,#10,fr_mousex,fr_mousey
	move.w		d0,fr_newobj		;save obj#
	cmp.w		fr_oldobj,d0		;same as last object?
	beq.s		.exit			;yup, do nothing
	bsr		fr_unhilite		;unhilite old obj if needed
	move.w		fr_newobj,fr_oldobj	;save new current obj#
	move.w		fr_newobj,d0		;get pointer to the obj
	cmpi.w		#-1,d0			;is pointer off form?
	beq.s		.exit			;yup, do nothing
	mulu		#24,d0			;finish getting pointer
	movea.l		fr_treeaddr,a0
	lea		(a0,d0.L),a0
	move.b		ob_type(a0),fr_extend	;save type for chks below!
	btst.b		#0,ob_state+1(a0)	;is it selected?
	bne.s		.nohl			;yep, go check passover bit
	btst.b		#0,fr_extend		;chk type hilite bit (8!)
	beq.s		.nohl			;not set, check passover bit
	bsr		fr_hilite
.nohl	btst.b		#1,fr_extend		;chk type passover bit (9!)
	beq.s		.exit			;not set, do nothing else
	move.b		#1,fr_exitnow		;show we wanna quit
	move.w		fr_newobj,d0		;now fix up our result
	bset.l		#PASSEXIT,d0		;flag a passover exit
	move.w		d0,fr_result		;and return it properly
.exit:	rts
*----------


*UNhilight an object, # passed in fr_hilited
*----------
fr_unhilite: tst.w	fr_hilited		;any hilited?
	beq.s		.exit			;nope, skip this!
	move.w		fr_hilited,d1		;show which object
	moveq		#0,d2			;flag to deselect it
	bsr.s		fr_redraw		;redraw it
	clr.w		fr_hilited		;show none hilited now
.exit:	rts
*----------


*hilight an object, # passed in fr_newobj
*----------
fr_hilite: move.w	fr_newobj,fr_hilited	;grab object number
	bne.s		.go			;skip if no object!
	rts
.go:	move.w		fr_hilited,d1		;show which object
	moveq		#1,d2			;flag to select it
*---------- FALLS THROUGH TO REDRAW ROUTINE! ----------*
*get an object's size/pos and state, then redraw (obj# passed in d1)
*d2.l = 0 to deselect or non-zero to select object
*----------
fr_redraw:	move.w	d1,fr_savobj		;save object number
	mulu		#24,d1			;get pointer to obj
	movea.l		fr_treeaddr,a0
	lea		(a0,d1.L),a0
	move.w		ob_state(a0),d0		;grab object state
	bclr.l		#0,d0			;assume we want deselect
	tst.l		d2			;did we want to select?
	beq.s		.go			;nope, clear it!
	bset.l		#0,d0			;select it
.go:	move.w		d0,fr_work		;save new state
	move.l		ob_width(a0),fr_w	;get w & h values
	objc_offset	fr_treeaddr,fr_savobj
	move.l		int_out+2,fr_x		;get x & y values
	objc_change	fr_treeaddr,fr_savobj,
&		fr_x,fr_y,fr_w,fr_h,fr_work,#1
.exit	rts
*----------


*-----------------------------------------------------------------------*
	section bss
*-----------------------------------------------------------------------*
fr_editobj	ds.w	1	;first editable object
fr_extras	ds.w	1	;flags: bit1=click-off exit, bit2=edit-wrap
fr_treeaddr	ds.l	1	;pointer to object tree
fr_cmdkeys	ds.l	1	;pointer to command key list
fr_equkeys	ds.l	1	;pointer to key equivalent list
fr_subkeys	ds.l	1	;pointer to key substitution list
fr_oldobj	ds.w	1	;previous selected object
fr_newobj	ds.w	1	;this (new) passover or clicked object
fr_hilited	ds.w	1	;hilited obj#, or 0 if none
fr_editidx	ds.w	1	;edit cursor index
fr_altstr	ds.l	1	;4-byte string for alt key digits
fr_result	ds.w	1	;value to return
fr_msgbuf	ds.w	16	;message buffer for evnt_multi
fr_happened	ds.w	1	;which event happened
fr_mousex	ds.w	1	;mouse x position
fr_mousey	ds.w	1	;mouse y position
fr_button	ds.w	1	;final button state
fr_shift	ds.w	1	;final shift key status
fr_keycode	ds.w	1	;keycode for pressed key
fr_clicked	ds.w	1	;number of clicks that occurred
fr_work		ds.w	1	;work area
fr_savobj	ds.w	1	;temp object# save area
fr_x		ds.w	1	;an object's x/y/w/h
fr_y		ds.w	1
fr_w		ds.w	1
fr_h		ds.w	1
fr_extend	ds.b	1	;BYTE, extended object type
fr_exitnow	ds.b	1	;BYTE flag, we want to exit
fr_passed	ds.b	1	;BYTE flag, walked past this edit object
*-----------------------------------------------------------------------*
	section text
